from apistar import http
from apistar.authentication import Authenticated
from apistar.types import Settings

from .exceptions import AuthenticationFailed
from .token import JWT


def get_jwt(authorization: http.Header, settings: Settings):
    if authorization is None:
        raise AuthenticationFailed('Authorization header is missing.') from None
    try:
        scheme, token = authorization.split()
    except ValueError:
        raise AuthenticationFailed('Could not seperate Authorization scheme and token.') from None
    if scheme.lower() != 'bearer':
        raise AuthenticationFailed('Authorization scheme not supported, try Bearer') from None
    return JWT(token=token, settings=settings)


class JWTAuthentication():
    def authenticate(self, authorization: http.Header, settings: Settings):
        jwt = get_jwt(authorization, settings)
        if jwt.payload == {}:
            raise AuthenticationFailed()
        jwt_settings = settings.get('JWT', {})
        uid = jwt.payload.get(jwt_settings.get('ID', 'id'), '')
        username = jwt.payload.get(jwt_settings.get('USERNAME', 'username'), '')
        return Authenticated(username, user={'id': uid, 'name': username}, token=jwt.token)



#!/usr/bin/env python
"""Tests for `apistar_jwt` package."""

import jwt
import pytest
from datetime import datetime, timedelta

from apistar import Component, Route, Settings, TestClient, exceptions, http, annotate
from apistar.interfaces import Auth
from apistar.frameworks.asyncio import ASyncIOApp
from apistar.frameworks.wsgi import WSGIApp
from apistar_jwt.authentication import JWTAuthentication, get_jwt
from apistar_jwt.token import JWT


# Test Routes
@annotate(authentication=[JWTAuthentication()])
def auth_required(request: http.Request, auth: Auth):
    return auth.user


def injected_component(request: http.Request, token: JWT):
    return token.payload


@pytest.mark.parametrize('app_class', [WSGIApp, ASyncIOApp])
def test_jwt_as_auth(app_class) -> None:
    routes = [
        Route('/auth-required-route', 'GET', auth_required),
    ]
    settings = {
        'JWT': {
            'SECRET': 'jwt-secret',
            'USERNAME': 'username',
            'ID': 'user',
        }
    }

    app = app_class(routes=routes, settings=settings)
    client = TestClient(app)

    response = client.get('/auth-required-route')
    assert response.status_code == 401

    payload = {'user': 1, 'username': 'bailey'}
    secret = settings['JWT']['SECRET']
    encoded_jwt = jwt.encode(payload, secret, algorithm='HS256').decode(encoding='UTF-8')

    response = client.get('/auth-required-route', headers={
        'Authorization': 'Bearer',
    })
    assert response.status_code == 401

    response = client.get('/auth-required-route', headers={
        'Authorization': 'Basic username',
    })
    assert response.status_code == 401

    response = client.get('/auth-required-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 200
    assert response.json() == {'id': payload[settings['JWT']['ID']], 'name': payload[settings['JWT']['USERNAME']]}  # noqa; E501

    encoded_jwt = jwt.encode(payload, 'wrong-secret').decode(encoding='UTF-8')
    response = client.get('/auth-required-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 401

    # wrong secret
    encoded_jwt = jwt.encode(payload, 'wrong secret', algorithm='HS256').decode(encoding='UTF-8')
    response = client.get('/auth-required-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 401

    # wrong algorithm
    encoded_jwt = jwt.encode(payload, secret, algorithm='HS512').decode(encoding='UTF-8')
    response = client.get('/auth-required-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 401

    # empty payload causes auth to fail
    encoded_jwt = jwt.encode({}, secret, algorithm='HS256').decode(encoding='UTF-8')
    response = client.get('/auth-required-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 401


@pytest.mark.parametrize('app_class', [WSGIApp, ASyncIOApp])
def test_jwt_issuer_claim(app_class) -> None:
    routes = [
        Route('/as-a-component-route', 'GET', injected_component),
        Route('/auth-required-route', 'GET', auth_required),
    ]
    settings = {
        'JWT': {
            'SECRET': 'jwt-secret',
            'USERNAME': 'username',
            'ID': 'user',
            'ISSUER': 'urn:foo',
        }
    }
    components = [
        Component(JWT, init=get_jwt)
    ]

    app = app_class(routes=routes, settings=settings, components=components)
    client = TestClient(app)

    payload = {'user': 1, 'username': 'bailey', 'iss': 'urn:foo'}
    secret = settings['JWT']['SECRET']

    # iss claim is correct
    encoded_jwt = jwt.encode(payload, secret, algorithm='HS256').decode(encoding='UTF-8')
    response = client.get('/auth-required-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 200
    assert response.json() == {'id': payload[settings['JWT']['ID']], 'name': payload[settings['JWT']['USERNAME']]}  # noqa; E501

    response = client.get('/as-a-component-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 200
    assert response.json() == payload

    # iss claim is incorrect
    payload['iss'] = 'urn:not-foo'
    encoded_jwt = jwt.encode(payload, secret, algorithm='HS256').decode(encoding='UTF-8')
    response = client.get('/auth-required-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 401

    response = client.get('/as-a-component-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 401

    # no iss claim included in jwt
    del payload['iss']
    encoded_jwt = jwt.encode(payload, secret, algorithm='HS256').decode(encoding='UTF-8')
    response = client.get('/auth-required-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 401

    response = client.get('/as-a-component-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 401


@pytest.mark.parametrize('app_class', [WSGIApp, ASyncIOApp])
def test_jwt_audience_claim(app_class) -> None:
    routes = [
        Route('/as-a-component-route', 'GET', injected_component),
        Route('/auth-required-route', 'GET', auth_required),
    ]
    settings = {
        'JWT': {
            'SECRET': 'jwt-secret',
            'USERNAME': 'username',
            'ID': 'user',
            'AUDIENCE': 'urn:foo',
        }
    }
    components = [
        Component(JWT, init=get_jwt)
    ]

    app = app_class(routes=routes, settings=settings, components=components)
    client = TestClient(app)

    payload = {'user': 1, 'username': 'bailey', 'aud': 'urn:foo'}
    secret = settings['JWT']['SECRET']

    # aud claim is single and correct
    encoded_jwt = jwt.encode(payload, secret, algorithm='HS256').decode(encoding='UTF-8')
    response = client.get('/auth-required-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 200
    assert response.json() == {'id': payload[settings['JWT']['ID']], 'name': payload[settings['JWT']['USERNAME']]}  # noqa; E501

    response = client.get('/as-a-component-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 200
    assert response.json() == payload

    # aud claim is multiple and correct
    payload['aud'] = ['urn:bar', 'urn:baz', 'urn:foo']
    encoded_jwt = jwt.encode(payload, secret, algorithm='HS256').decode(encoding='UTF-8')
    response = client.get('/auth-required-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 200
    assert response.json() == {'id': payload[settings['JWT']['ID']], 'name': payload[settings['JWT']['USERNAME']]}  # noqa; E501

    response = client.get('/as-a-component-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 200
    assert response.json() == payload

    # aud claim is incorrect and single
    payload['aud'] = 'urn:not-foo'
    encoded_jwt = jwt.encode(payload, secret, algorithm='HS256').decode(encoding='UTF-8')
    response = client.get('/auth-required-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 401

    response = client.get('/as-a-component-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 401

    # aud claim is incorrect and multiple
    payload['aud'] = ['urn:bar', 'urn:baz', 'urn:not-foo']
    encoded_jwt = jwt.encode(payload, secret, algorithm='HS256').decode(encoding='UTF-8')
    response = client.get('/auth-required-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 401

    response = client.get('/as-a-component-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 401

    # no aud claim included in jwt
    del payload['aud']
    encoded_jwt = jwt.encode(payload, secret, algorithm='HS256').decode(encoding='UTF-8')
    response = client.get('/auth-required-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 401

    response = client.get('/as-a-component-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 401


@pytest.mark.parametrize('app_class', [WSGIApp, ASyncIOApp])
def test_jwt_leeway_claim(app_class) -> None:
    routes = [
        Route('/as-a-component-route', 'GET', injected_component),
        Route('/auth-required-route', 'GET', auth_required),
    ]
    settings = {
        'JWT': {
            'SECRET': 'jwt-secret',
            'USERNAME': 'username',
            'ID': 'user',
            'LEEWAY': 3,
        }
    }
    components = [
        Component(JWT, init=get_jwt)
    ]

    app = app_class(routes=routes, settings=settings, components=components)
    client = TestClient(app)

    payload = {'user': 1, 'username': 'bailey', 'exp': datetime.utcnow() - timedelta(seconds=2)}
    secret = settings['JWT']['SECRET']

    # exp claim doesn't fail because of leeway
    encoded_jwt = jwt.encode(payload, secret, algorithm='HS256').decode(encoding='UTF-8')
    response = client.get('/auth-required-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 200
    assert response.json() == {'id': payload[settings['JWT']['ID']], 'name': payload[settings['JWT']['USERNAME']]}  # noqa; E501

    response = client.get('/as-a-component-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 200
    assert response.json() == payload

    # exp claim fails because leeway is only 3 seconds
    payload['exp'] = datetime.utcnow() - timedelta(seconds=4)
    encoded_jwt = jwt.encode(payload, secret, algorithm='HS256').decode(encoding='UTF-8')
    response = client.get('/auth-required-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 401

    response = client.get('/as-a-component-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 401

    # no exp claim included in jwt, leeway doesnt apply
    del payload['exp']
    encoded_jwt = jwt.encode(payload, secret, algorithm='HS256').decode(encoding='UTF-8')
    response = client.get('/auth-required-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 200

    response = client.get('/as-a-component-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 200


@pytest.mark.parametrize('app_class', [WSGIApp, ASyncIOApp])
def test_jwt_as_component(app_class) -> None:
    routes = [
        Route('/as-a-component-route', 'GET', injected_component),
    ]
    settings = {
        'JWT': {
            'SECRET': 'jwt-secret',
            'USERNAME': 'username',
            'ID': 'user',
        }
    }
    components = [
        Component(JWT, init=get_jwt)
    ]

    app = app_class(routes=routes, settings=settings, components=components)
    client = TestClient(app)

    payload = {'user': 1, 'username': 'bailey'}
    secret = settings['JWT']['SECRET']
    encoded_jwt = jwt.encode(payload, secret, algorithm='HS256').decode(encoding='UTF-8')

    response = client.get('/as-a-component-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 200
    assert response.json() == payload

    response = client.get('/as-a-component-route', headers={
        'Authorization': 'Bearer ',
    })
    assert response.status_code == 401

    # wrong secret
    encoded_jwt = jwt.encode(payload, 'wrong secret', algorithm='HS256').decode(encoding='UTF-8')
    response = client.get('/as-a-component-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 401

    # wrong algorithm
    encoded_jwt = jwt.encode(payload, secret, algorithm='HS512').decode(encoding='UTF-8')
    response = client.get('/as-a-component-route', headers={
        'Authorization': 'Bearer {token}'.format(token=encoded_jwt),
    })
    assert response.status_code == 401


def test_jwt_encode() -> None:
    payload = {'email': 'test@example.com'}
    token = jwt.encode(payload, 'jwt-secret', algorithm='HS256').decode(encoding='UTF-8')
    secret = 'jwt-secret'
    encoded_jwt = JWT.encode(payload=payload, secret=secret)
    assert encoded_jwt == token

    encoded_jwt = JWT.encode(payload=payload, secret=secret, algorithm='HS512')
    assert encoded_jwt != token
    token = jwt.encode(payload, 'jwt-secret', algorithm='HS512').decode(encoding='UTF-8')
    assert encoded_jwt == token


def test_misconfigured_jwt_settings() -> None:
    settings = Settings({
        'JWT': {},
    })
    token = 'abc'

    with pytest.raises(exceptions.ConfigurationError):
        JWT(token=token, settings=settings)


def test_no_secret_passed_to_encode() -> None:
    payload = {'some': 'payload'}
    with pytest.raises(exceptions.ConfigurationError):
        JWT.encode(payload=payload)


def test_unknown_algorithm_passed_to_encode() -> None:
    payload = {'some': 'payload'}
    with pytest.raises(exceptions.ConfigurationError):
        JWT.encode(payload=payload, secret='jwt-secret', algorithm='unknown-algorithm')



from apistar.exceptions import ConfigurationError
from apistar.types import Settings
import jwt

from .exceptions import AuthenticationFailed


class JWT():
    def __init__(self, settings: Settings, token=None, issuer=None, audience=None, leeway=None):
        jwt_settings = settings.get('JWT', {})
        self.secret = jwt_settings.get('SECRET', None)
        if self.secret is None:
            msg = 'The SECRET setting under JWT settings must be defined.'
            raise ConfigurationError(msg) from None
        self.algorithms = jwt_settings.get('ALGORITHMS', ['HS256'])
        self.token = token
        self.issuer = issuer if issuer is not None else jwt_settings.get('ISSUER', None)
        self.audience = audience if audience is not None else jwt_settings.get('AUDIENCE', None)
        self.leeway = leeway if leeway is not None else jwt_settings.get('LEEWAY', None)
        try:
            kwargs = {}
            if self.issuer:
                kwargs.update({'issuer': self.issuer})
            if self.audience:
                kwargs.update({'audience': self.audience})
            if self.leeway:
                kwargs.update({'leeway': self.leeway})
            self.payload = jwt.decode(
                self.token, self.secret, algorithms=self.algorithms, **kwargs)
        except Exception:
            self.payload = None
            raise AuthenticationFailed()

    @staticmethod
    def encode(payload, secret=None, algorithm=None, **kwargs):
        if secret is None:
            msg = 'The secret keyword argument must be defined.'
            raise ConfigurationError(msg) from None
        algorithm = 'HS256' if algorithm is None else algorithm
        try:
            token = jwt.encode(
                payload, secret, algorithm=algorithm, **kwargs).decode(encoding='UTF-8')
        except Exception as exc:
            raise ConfigurationError(exc.__class__.__name__) from None
        return token
